@@ -4,8 +4,9 @@ import re
 from datetime import date, datetime
 from functools import reduce
 from math import floor
+from operator import truth
 from string import ascii_lowercase, digits
-from typing import Any, Dict, List, Optional, Pattern
+from typing import Any, Dict, List, Optional, Pattern, Union
 from unicodedata import normalize
 
 from beets import __version__ as beets_version
@@ -41,17 +42,14 @@ MEDIA_MAP = {
 VALID_URL_CHARS = {*ascii_lowercase, *digits}
 
 _catalognum = r"([^\d\W]+[_\W]?\d{2,}(?:\W\d|CD)?)"
+_exclusive = r"\s?[\[(](bandcamp )?(digi(tal)? )?(bonus|only|exclusive)[\])]"
 PATTERNS: Dict[str, Pattern] = {
     "meta": re.compile(r".*datePublished.*", flags=re.MULTILINE),
     "quick_catalognum": re.compile(rf"\[{_catalognum}\]"),
     "catalognum": re.compile(rf"^{_catalognum}|{_catalognum}$"),
-    "catalognum_exclude": re.compile(
-        r"vol(ume)?|artists?|2020|2021|(^|\s)C\d\d|\d+/\d+", flags=re.IGNORECASE
-    ),
-    "country": re.compile(r'location\ssecondaryText">(?:[\w\s]*, )?([\w\s,]+){1,4}'),
-    "digital": re.compile(
-        r"(^DIGI ([\d]+\.\s?)?)|(?i:\s?[\[(].*digi(tal)? (bonus|only|exclusive)[\])])"
-    ),
+    "catalognum_excl": re.compile(r"(?i:vol(ume)?|artists)|202[01]|(^|\s)C\d\d|\d+/\d+"),
+    "country": re.compile(r'location\ssecondaryText">(?:[\w\s.]*, )?([\w\s,]+){1,4}'),
+    "digital": re.compile(rf"^DIGI (\d+\.\s?)?|(?i:{_exclusive})"),
     "label": re.compile(r'og:site_name".*content="([^"]*)"'),
     "lyrics": re.compile(r'"lyrics":({[^}]*})'),
     "release_date": re.compile(r"release[ds] ([\d]{2} [A-Z][a-z]+ [\d]{4})"),
@@ -91,21 +89,20 @@ class Helpers:
         count: str = match.groupdict()["count"]
         return int(count) if count.isdigit() else conv[count.lower()]
 
+    @staticmethod
+    def check_digital_only(name: str) -> Dict[str, Union[bool, str]]:
+        no_digi_only_name = re.sub(PATTERNS["digital"], "", name)
+        if no_digi_only_name != name:
+            return dict(digital_only=True, name=no_digi_only_name)
+        return dict(digital_only=False)
+
     @staticmethod
     def parse_track_name(name: str) -> JSONDict:
         match = re.search(PATTERNS["track_name"], name)
         try:
-            track = match.groupdict()
+            return match.groupdict()  # type: ignore
         except AttributeError:
-            track = {"title": name, "artist": None, "track_alt": None}
-        title = track["title"]
-        excl_digi_only_title = re.sub(PATTERNS["digital"], "", title)
-        if excl_digi_only_title != title:
-            track["digital_only"] = True
-            track["title"] = excl_digi_only_title
-        else:
-            track["digital_only"] = False
-        return track
+            return {"title": name, "artist": None, "track_alt": None}
 
     @staticmethod
     def parse_catalognum(album: str, disctitle: str) -> str:
@@ -114,7 +111,7 @@ class Helpers:
             (PATTERNS["catalognum"], disctitle),
             (PATTERNS["catalognum"], album),
         ]:
-            match = re.search(pattern, re.sub(PATTERNS["catalognum_exclude"], "", string))
+            match = re.search(pattern, re.sub(PATTERNS["catalognum_excl"], "", string))
             if match:
                 return [group for group in match.groups() if group].pop()
 
@@ -132,6 +129,22 @@ class Helpers:
                 return floor(item.get("value", 0))
         return 0
 
+    @staticmethod
+    def clean_up_album_name(name: str, *args) -> str:
+        excl = "Various Artists|limited edition"
+
+        # remove special chars
+        special = r"[*.+\\]"
+        extras = "|".join(filter(truth, map(lambda x: re.sub(special, "", x), args)))
+
+        excl = rf"({excl}|{extras})" if extras else rf"({excl})"
+
+        pat = re.compile(
+            rf" EP|({excl}\s[|/]+\s?)|[\[(]{excl}[])]|(\s-\s){excl}|{excl}(\s?-\s)",
+            flags=re.IGNORECASE,
+        )
+        return re.sub(pat, "", name).strip()
+
 
 class Metaguru(Helpers):
     html: str
@@ -139,11 +152,13 @@ class Metaguru(Helpers):
     meta: JSONDict
 
     _media = None  # type: Optional[Dict[str, str]]
+    _singleton = False
 
     def __init__(self, html: str, media: str = DEFAULT_MEDIA) -> None:
         self.html = html
         self.preferred_media = media
 
+        self.meta = {}
         match = re.search(PATTERNS["meta"], html)
         if match:
             self.meta = json.loads(match.group())
@@ -152,11 +167,17 @@ class Metaguru(Helpers):
         match = re.search(pattern, self.html)
         return match.groups()[0] if match else ""
 
-    @property
+    @cached_property
     def album_name(self) -> str:
-        # TODO: Cleanup catalogue, etc
         return self.meta["name"]
 
+    @cached_property
+    def clean_album_name(self) -> str:
+        args = [self.label, self.catalognum]
+        if not self._singleton:
+            args.append(self.albumartist)
+        return self.clean_up_album_name(self.album_name, *args)
+
     @property
     def album_id(self) -> str:
         return self.meta["@id"]
@@ -171,7 +192,7 @@ class Metaguru(Helpers):
         image = self.meta.get("image", "")
         return image[0] if isinstance(image, list) else image
 
-    @property
+    @cached_property
     def label(self) -> str:
         return self._search(PATTERNS["label"])
 
@@ -188,7 +209,7 @@ class Metaguru(Helpers):
         datestr = self.parse_release_date(self.html)
         return datetime.strptime(datestr, DATE_FORMAT).date()
 
-    @property
+    @cached_property
     def disctitle(self) -> str:
         if self._media and self.media != "Digital Media":
             return self._media.get("name", "")
@@ -231,6 +252,15 @@ class Metaguru(Helpers):
         """We can't tell the number of current disc for a multi-disc release."""
         return 1
 
+    @property
+    def description(self) -> str:
+        descr = self.meta.get("description", "")
+        if not descr and self._media:
+            descr = self._media.get("description", "")
+            if descr.startswith("Includes high-quality dow"):
+                descr = ""
+        return descr
+
     @cached_property
     def tracks(self) -> List[JSONDict]:
         """`raw_track` example
@@ -252,9 +282,14 @@ class Metaguru(Helpers):
         },
         """
         tracks = []
-        for raw_track in self.meta["track"].get("itemListElement", []):
+        if not self._singleton:
+            raw_tracks = self.meta["track"].get("itemListElement", [])
+        else:
+            raw_tracks = [{"item": self.meta}]
+        for raw_track in raw_tracks:
             track = raw_track["item"]
-            track["position"] = raw_track["position"]
+            track["position"] = raw_track.get("position") or 1
+            track.update(self.check_digital_only(track["name"]))
             track.update(self.parse_track_name(track["name"]))
             tracks.append(track)
         return tracks
@@ -266,19 +301,25 @@ class Metaguru(Helpers):
             for t in self.tracks
         }
         unique_artists.discard(None)
-        if any(
-            ("EP" in self.disctitle, "EP" in self.album_name, len(unique_artists) <= 1)
-        ):
-            return True
-        return False
+        return len(unique_artists) <= 1
+
+    @property
+    def is_lp(self) -> bool:
+        return "LP" in self.album_name or "LP" in self.disctitle
+
+    @cached_property
+    def is_ep(self) -> bool:
+        return (
+            "EP" in self.album_name
+            or "EP" in self.disctitle
+            or (self.media == "Vinyl" and len(self.tracks) == 4)
+        )
 
     @cached_property
     def is_va(self) -> bool:
-        if "Various Artists" in self.album_name or (
-            len(self.tracks) > 4 and not self.is_single_artist
-        ):
-            return True
-        return False
+        return "various artists" in self.album_name.lower() or (
+            not self.is_single_artist and not self.is_ep
+        )
 
     @property
     def bandcamp_albumartist(self) -> str:
@@ -294,49 +335,62 @@ class Metaguru(Helpers):
 
     @property
     def albumtype(self) -> str:
-        if "LP" in self.album_name or "LP" in self.disctitle:
+        if self._singleton:
+            return "single"
+        if self.is_lp:
             return "album"
         if self.is_va:
             return "compilation"
-        if self.catalognum or "EP" in self.disctitle or "EP" in self.album_name:
+        if self.is_ep:
             return "ep"
         return "album"
 
-    def _trackinfo(self, track: JSONDict, medium_total: int) -> TrackInfo:
-        data = {
-            "artist": track.get("artist") or self.albumartist,
-            "artist_id": self.artist_id,
-            "data_source": DATA_SOURCE,
-            "data_url": self.album_id,
-            "index": track.get("position"),
-            "length": self.get_duration(track),
-            "media": self.media or DEFAULT_MEDIA,
-            "track_alt": track.get("track_alt"),
-            "disctitle": self.disctitle,
-            "medium": self.medium,
-            "medium_index": track.get("position"),
-            "medium_total": medium_total,
-        }
-        if NEW_BEETS:
-            return TrackInfo(title=track.get("title"), track_id=track.get("url"), **data)
-        return TrackInfo(track.get("title"), track.get("url"), **data)
+    @property
+    def _common(self) -> JSONDict:
+        return dict(
+            data_source=DATA_SOURCE,
+            media=self.media or DEFAULT_MEDIA,
+            data_url=self.album_id,
+            artist_id=self.artist_id,
+        )
+
+    @property
+    def _common_album(self) -> JSONDict:
+        return dict(
+            year=self.release_date.year,
+            month=self.release_date.month,
+            day=self.release_date.day,
+            label=self.label,
+            catalognum=self.catalognum,
+            albumtype=self.albumtype,
+            album=self.clean_album_name,
+            albumstatus=OFFICIAL if self.release_date < date.today() else PROMO,
+            country=self.country,
+        )
+
+    def _trackinfo(self, track: JSONDict, medium_total: int, **kwargs) -> TrackInfo:
+        return TrackInfo(
+            title=track.get("title"),
+            track_id=track.get("url"),
+            artist=track.get("artist") or self.albumartist,
+            index=track.get("position"),
+            length=self.get_duration(track),
+            track_alt=track.get("track_alt"),
+            disctitle=self.disctitle or None,
+            medium=self.medium,
+            medium_index=track.get("position"),
+            medium_total=medium_total,
+            **self._common,
+            **kwargs,
+        )
 
     @property
     def singleton(self) -> TrackInfo:
-        track = self.parse_track_name(self.album_name)
-        data = {
-            "artist": track.get("artist") or self.bandcamp_albumartist,
-            "artist_id": self.artist_id,
-            "data_source": DATA_SOURCE,
-            "data_url": self.album_id,
-            "index": 1,
-            "length": self.get_duration(self.meta),
-            "media": self.media or DEFAULT_MEDIA,
-            "track_alt": track.get("track_alt"),
-        }
+        self._singleton = True
+        kwargs: JSONDict = {}
         if NEW_BEETS:
-            return TrackInfo(title=track.get("title"), track_id=self.album_id, **data)
-        return TrackInfo(track.get("title"), self.album_id, **data)
+            kwargs.update(**self._common_album, albumartist=self.bandcamp_albumartist)
+        return self._trackinfo(self.tracks.pop(), 1, **kwargs)
 
     def albuminfo(self, include_all: bool) -> AlbumInfo:
         if self.media == "Digital Media" or include_all:
@@ -346,44 +400,19 @@ class Metaguru(Helpers):
 
         medium_total = len(filtered_tracks)
         _tracks = [self._trackinfo(track, medium_total) for track in filtered_tracks]
-        rel_date = self.release_date
-        data = {
-            "va": self.is_va,
-            "year": rel_date.year,
-            "month": rel_date.month,
-            "day": rel_date.day,
-            "label": self.label,
-            "catalognum": self.catalognum,
-            "albumtype": self.albumtype,
-            "data_url": self.album_id,
-            "albumstatus": OFFICIAL if rel_date < date.today() else PROMO,
-            "country": self.country,
-            "media": self.media,
-            "mediums": self.mediums,
-            "data_source": DATA_SOURCE,
-        }
-        if not NEW_BEETS:
-            return AlbumInfo(
-                self.album_name,
-                self.album_id,
-                self.albumartist,
-                self.artist_id,
-                tracks=_tracks,
-                **data,
-            )
         return AlbumInfo(
-            _tracks,
-            **{
-                "album": self.album_name,
-                "albumartist": self.albumartist,
-                "album_id": self.album_id,
-                "artist_id": self.artist_id,
-            },
-            **data,
+            **self._common,
+            **self._common_album,
+            artist=self.albumartist,
+            album_id=self.album_id,
+            va=self.is_va,
+            mediums=self.mediums,
+            tracks=_tracks,
         )
 
     def album(self, include_all: bool) -> AlbumInfo:
         """Return album for the appropriate release format."""
+        # map available formats to appropriate names
         medias: JSONDict = {}
         try:
             for _format in self.meta["albumRelease"]:
@@ -395,10 +424,12 @@ class Metaguru(Helpers):
         except (KeyError, AttributeError):
             return None
 
+        # if preference is given and the format is available, return it
         for preference in self.preferred_media.split(","):
             if preference in medias:
                 self._media = medias[preference]
-                return self.albuminfo(include_all)
+                break
+        else:  # otherwise, use the default option
+            self._media = medias[DEFAULT_MEDIA]
 
-        self._media = medias[DEFAULT_MEDIA]
         return self.albuminfo(include_all)
